import IPython.nbformat.current as current


class NotebookParser(object):
    """Parser for IPython Notebook files."""
    def parse(self, json_data):
        """Parse a notebook .ipynb file.

        Parameters
        ----------
        json_data : file
            A file handle for an .ipynb file.

        Returns
        -------
        nb : An IPython Notebook data structure.
        """
        data = current.read(json_data, 'ipynb')
        json_data.close()
        return data

    # param:
    # json_data_string: raw unicode string
    def parseString(self, json_data_string):
        return current.reads(json_data_string, 'ipynb')



'''
Entry points for the nbdiff package.
'''
from __future__ import print_function
import argparse
from .merge import notebook_merge
from .notebook_parser import NotebookParser
import sys
from .notebook_diff import notebook_diff
from .adapter.git_adapter import GitAdapter
from .server.local_server import app
import threading
import webbrowser
import IPython.nbformat.current as nbformat


def diff():
    description = '''
    Produce a diffed IPython Notebook from before and after notebooks.

    If no arguments are given, nbdiff looks for modified notebook files in
    the version control system.

    The resulting diff is presented to the user in the browser at
    http://localhost:5000.
    '''
    usage = 'nbdiff [-h] [--check] [--browser=<browser] [before after]'
    parser = argparse.ArgumentParser(
        description=description,
        usage=usage,
    )
    # TODO share this code with merge()
    parser.add_argument(
        '--browser',
        '-b',
        default=None,
        help='Browser to launch nbdiff/nbmerge in',
    )
    parser.add_argument(
        '--check',
        '-c',
        action='store_true',
        default=False,
        help='Run nbdiff algorithm but do not display the result.',
    )
    parser.add_argument('before', nargs='?',
                        help='The notebook to diff against.')
    parser.add_argument('after', nargs='?',
                        help='The notebook to compare `before` to.')
    args = parser.parse_args()

    parser = NotebookParser()

    if args.before and args.after:
        invalid_notebooks = []

        try:
            notebook1 = parser.parse(open(args.before))
        except nbformat.NotJSONError:
            invalid_notebooks.append(args.before)

        try:
            notebook2 = parser.parse(open(args.after))
        except nbformat.NotJSONError:
            invalid_notebooks.append(args.after)

        if (len(invalid_notebooks) == 0):
            result = notebook_diff(notebook1, notebook2)

            filename_placeholder = "{} and {}".format(args.before, args.after)
            app.add_notebook(result, filename_placeholder)
            if not args.check:
                open_browser(args.browser)
                app.run(debug=False)

        else:
            print('The notebooks could not be diffed.')
            print('There was a problem parsing the following notebook '
                  + 'files:\n' + '\n'.join(invalid_notebooks))
            return -1

    elif not (args.before or args.after):
        # No arguments have been given. Ask version control instead
        git = GitAdapter()

        modified_notebooks = git.get_modified_notebooks()

        if not len(modified_notebooks) == 0:
            invalid_notebooks = []
            for nbook in modified_notebooks:
                try:
                    filename = nbook[2]

                    current_notebook = parser.parse(nbook[0])
                    head_version = parser.parse(nbook[1])

                    result = notebook_diff(head_version, current_notebook)
                    app.add_notebook(result, filename)

                except nbformat.NotJSONError:
                    invalid_notebooks.append(filename)

            if (len(invalid_notebooks) > 0):
                print('There was a problem parsing the following notebook '
                      + 'files:\n' + '\n'.join(invalid_notebooks))

            if (len(modified_notebooks) == len(invalid_notebooks)):
                print("There are no valid notebooks to diff.")
                return -1

            if not args.check:
                open_browser(args.browser)
                app.run(debug=False)
        else:
            print("No modified files to diff.")
            return 0
    else:
        print ("Invalid number of arguments. Run nbdiff --help")
        return -1


def merge():
    description = '''
    nbmerge is a tool for resolving merge conflicts in IPython Notebook
    files.

    If no arguments are given, nbmerge attempts to find the conflicting
    file in the version control system.

    Positional arguments are available for integration with version
    control systems such as Git and Mercurial.
    '''
    usage = (
        'nbmerge [-h] [--check] [--browser=<browser>]'
        '[local base remote [result]]'
    )
    parser = argparse.ArgumentParser(
        description=description,
        usage=usage,
    )
    parser.add_argument('notebook', nargs='*')
    # TODO share this code with diff()
    parser.add_argument(
        '--check',
        '-c',
        action='store_true',
        default=False,
        help='Run nbmerge algorithm but do not display the result.',
    )
    parser.add_argument(
        '--browser',
        '-b',
        default=None,
        help='Browser to launch nbdiff/nbmerge in',
    )
    args = parser.parse_args()
    length = len(args.notebook)
    parser = NotebookParser()
    valid_notebooks = False

    # only 'nbmerge' - no files specified with command
    if length == 0:
        git = GitAdapter()
        unmerged_notebooks = git.get_unmerged_notebooks()

        if not len(unmerged_notebooks) == 0:
            invalid_notebooks = []

            for nbook in unmerged_notebooks:
                try:
                    filename = nbook[3]

                    nb_local = parser.parse(nbook[0])
                    nb_base = parser.parse(nbook[1])
                    nb_remote = parser.parse(nbook[2])

                    pre_merged_notebook = notebook_merge(nb_local,
                                                         nb_base, nb_remote)
                    app.add_notebook(pre_merged_notebook, filename)

                except nbformat.NotJSONError:
                    invalid_notebooks.append(filename)

            if (len(invalid_notebooks) > 0):
                print('There was a problem parsing the following notebook '
                      + 'files:\n' + '\n'.join(invalid_notebooks))

            if (len(unmerged_notebooks) == len(invalid_notebooks)):
                print("There are no valid notebooks to merge.")
                return -1
            else:
                valid_notebooks = True

        else:
            print('There are no files to be merged.')
            return -1

    # files specified with nbmerge command
    elif length == 3 or length == 4:
        invalid_notebooks = []

        if length == 3:
            # hg usage:
            # $ hg merge -t nbmerge <branch>

            # Mercurial gives three arguments:
            # 1. Local / Result (the file in your working directory)
            # 2. Base
            # 3. Remote
            filename = args.notebook[0]  # filename for saving

        elif length == 4:
            # You need to run this before git mergetool will accept nbmerge
            # $ git config mergetool.nbmerge.cmd \
            #        "nbmerge \$LOCAL \$BASE \$REMOTE \$MERGED"
            # and then you can invoke it with:
            # $ git mergetool -t nbmerge
            #
            # Git gives four arguments (these are configurable):
            # 1. Local
            # 2. Base
            # 3. Remote
            # 4. Result (the file in your working directory)
            filename = args.notebook[3]   # filename for saving

        try:
            nb_local = parser.parse(open(args.notebook[0]))
        except nbformat.NotJSONError:
            invalid_notebooks.append(args.notebook[0])

        try:
            nb_base = parser.parse(open(args.notebook[1]))
        except nbformat.NotJSONError:
            invalid_notebooks.append(args.notebook[1])

        try:
            nb_remote = parser.parse(open(args.notebook[2]))
        except nbformat.NotJSONError:
            invalid_notebooks.append(args.notebook[2])

        # local, base and remote are all valid notebooks
        if (len(invalid_notebooks) == 0):
            pre_merged_notebook = notebook_merge(nb_local, nb_base, nb_remote)
            app.add_notebook(pre_merged_notebook, filename)
            valid_notebooks = True

        elif (len(invalid_notebooks) > 0):
                print('There was a problem parsing the following notebook '
                      + 'files:\n' + '\n'.join(invalid_notebooks))
                print("There are no valid notebooks to merge.")
                return -1

    else:
        sys.stderr.write('Incorrect number of arguments. Quitting.\n')
        sys.exit(-1)

    def save_notebook(notebook_result, filename):
        parsed = nbformat.reads(notebook_result, 'json')
        with open(filename, 'w') as targetfile:
            nbformat.write(parsed, targetfile, 'ipynb')

    if (valid_notebooks):
        if not args.check:
            app.shutdown_callback(save_notebook)
            open_browser(args.browser)
            app.run(debug=False)


def open_browser(browser_exe):
    try:
        browser = webbrowser.get(browser_exe)
    except webbrowser.Error:
        browser = None
    if browser:
        b = lambda: browser.open("http://127.0.0.1:5000/0", new=2)
        threading.Thread(target=b).start()



from .diff import (
    diff,
    create_grid,
    find_matches,
)
import Levenshtein


class BooleanPlus(object):

    def __init__(self, truthfulness, mod):
        self.truth = truthfulness
        self.modified = mod

    def __nonzero__(self):
        '''
        for evaluating as a boolean
        '''
        return self.truth

    def is_modified(self):
        return self.modified


class LineComparator(object):

    def __init__(self, data, check_modified=False):
        self.data = data
        self.check_modified = check_modified

    def __eq__(self, other):
        return self.equal(self.data, other.data)

    def equal(self, line1, line2):
        '''
        return true if exactly equal or if equal but modified,
        otherwise return false
        return type: BooleanPlus
        '''

        eqLine = line1 == line2
        if eqLine:
            return BooleanPlus(True, False)
        else:
            unchanged_count = self.count_similar_words(line1, line2)
            similarity_percent = (
                (2.0 * unchanged_count) /
                (len(line1.split()) + len(line2.split()))
            )
            if similarity_percent >= 0.50:
                return BooleanPlus(True, True)
            return BooleanPlus(False, False)

    def count_similar_words(self, line1, line2):
        words1 = line1.split()
        words2 = line2.split()
        grid = create_grid(words1, words2)
        matches = []

        for colnum in range(len(grid)):
            new_matches = find_matches(grid[colnum], colnum)
            matches = matches + new_matches

        matched_cols = [r[0] for r in matches]
        matched_rows = [r[1] for r in matches]
        unique_cols = set(matched_cols)
        unique_rows = set(matched_rows)

        return min(len(unique_cols), len(unique_rows))


class CellComparator():

    def __init__(self, data, check_modified=False):
        self.data = data
        self.check_modified = check_modified

    def __eq__(self, other):
        return self.equal(self.data, other.data)

    def equal(self, cell1, cell2):
        if not cell1["cell_type"] == cell2["cell_type"]:
            return False
        elif cell1["cell_type"] == "heading":
            if not cell1["level"] == cell2["level"]:
                return False
            else:
                if cell1['source'] == cell2['source']:
                    return True
                result = diff(
                    cell1["source"].split(' '),
                    cell2["source"].split(' ')
                )
                modified = 0.0
                unchanged = 0.0
                for dict in result:
                    if dict['state'] == "added" or dict['state'] == "deleted":
                        modified += 1
                    elif dict['state'] == "unchanged":
                        unchanged += 1
                modifiedness = unchanged/(modified + unchanged)
                if modifiedness >= 0.6:
                    return BooleanPlus(True, True)
                else:
                    return False
        elif self.istextcell(cell1):
            return cell1["source"] == cell2["source"]
        else:
            return self.compare_cells(cell1, cell2)

    def istextcell(self, cell):
        return "source" in cell

    def equaloutputs(self, output1, output2):
        if not len(output1) == len(output2):
            return False
        for i in range(0, len(output1)):
            if not output1[i] == output2[i]:
                return False
        return True

    def compare_cells(self, cell1, cell2):
        '''
        return true if exactly equal or if equal but modified,
        otherwise return false
        return type: BooleanPlus
        '''
        eqlanguage = cell1["language"] == cell2["language"]
        eqinput = cell1["input"] == cell2["input"]
        eqoutputs = self.equaloutputs(cell1["outputs"], cell2["outputs"])

        if eqlanguage and eqinput and eqoutputs:
            return BooleanPlus(True, False)
        elif not self.check_modified:
            return BooleanPlus(False, False)

        input1 = u"".join(cell1['input'])
        input2 = u"".join(cell2['input'])
        similarity_percent = Levenshtein.ratio(input1, input2)
        if similarity_percent >= 0.65:
            return BooleanPlus(True, True)
        return BooleanPlus(False, False)



__author__ = 'root'

import sys
import subprocess
from .vcs_adapter import VcsAdapter


class GitAdapter(VcsAdapter):

    def get_modified_notebooks(self):
        # get modified file names
        modified = subprocess.check_output("git ls-files --modified".split())
        fnames = modified.splitlines()

        # get unmerged file info
        unmerged = subprocess.check_output("git ls-files --unmerged".split())
        unmerged_array = [line.split() for line in unmerged.splitlines()]

        # get unmerged file names
        unmerged_array_names = [x[3] for x in unmerged_array]

        # ignore unmerged files, get unique names
        fnames = list(set(fnames) - set(unmerged_array_names))

        nb_diff = []
        for item in fnames:
            head_version_show = subprocess.Popen(
                ['git', 'show', 'HEAD:' + item],
                stdout=subprocess.PIPE
            )

            current_local_notebook = open(item)
            committed_notebook = head_version_show.stdout

            nb_diff.append((current_local_notebook, committed_notebook, item))

        return super(GitAdapter, self).filter_modified_notebooks(nb_diff)

    def get_unmerged_notebooks(self):
        # TODO error handling.

        output = subprocess.check_output("git ls-files --unmerged".split())
        output_array = [line.split() for line in output.splitlines()]

        if len(output_array) % 3 != 0:  # TODO should be something else
            sys.stderr.write(
                "Can't find the conflicting notebook. Quitting.\n")
            sys.exit(-1)

        hash_list = []

        for index in xrange(0, len(output_array), 3):
            local_hash = output_array[index + 1][1]
            base_hash = output_array[index][1]
            remote_hash = output_array[index + 2][1]
            file_name = output_array[index][3]
            hash_list.append((local_hash, base_hash, remote_hash, file_name))

        file_hooks = []

        for hash in hash_list:
            local = subprocess.Popen(
                ['git', 'show', hash[0]],
                stdout=subprocess.PIPE
            )
            base = subprocess.Popen(
                ['git', 'show', hash[1]],
                stdout=subprocess.PIPE
            )
            remote = subprocess.Popen(
                ['git', 'show', hash[2]],
                stdout=subprocess.PIPE
            )
            file_name = hash[3]
            file_hooks.append((local.stdout, base.stdout,
                              remote.stdout, file_name))

        return super(GitAdapter, self).filter_unmerged_notebooks(file_hooks)

    def stage_file(self, file, contents=None):
        if contents is not None:
            with open(file, 'w') as result_file:
                result_file.write(file)
        command = ["git", "add", file]
        return subprocess.call(command)



__author__ = 'root'


class HgAdapter:

    def get_modified_notebooks(self):
        pass

    def get_unmerged_notebooks(self):
        pass

    def stage_file(self, file, contents=None):
        pass



__author__ = 'root'

import re


class VcsAdapter(object):

    def get_modified_notebooks(self):
        raise NotImplementedError("Subclass must implement abstract method")

    def filter_modified_notebooks(self, file_hooks):
        modified_notebooks = []
        for item in file_hooks:
            if re.search('.ipynb$', item[2]):
                modified_notebooks.append(item)

        return modified_notebooks

    def get_unmerged_notebooks(self):
        raise NotImplementedError("Subclass must implement abstract method")

    def filter_unmerged_notebooks(self, file_hooks):
        unmerged_notebooks = []
        for item in file_hooks:
            if re.search('.ipynb$', item[3]):
                unmerged_notebooks.append(item)

        return unmerged_notebooks

    def stage_file(self, file, contents=None):
        raise NotImplementedError("Subclass must implement abstract method")



import itertools as it
import collections

__all__ = ['diff']


def diff(before, after, check_modified=False):
    """Diff two sequences of comparable objects.

    The result of this function is a list of dictionaries containing
    values in ``before`` or ``after`` with a ``state`` of either
    'unchanged', 'added', 'deleted', or 'modified'.

    >>> import pprint
    >>> result = diff(['a', 'b', 'c'], ['b', 'c', 'd'])
    >>> pprint.pprint(result)
    [{'state': 'deleted', 'value': 'a'},
     {'state': 'unchanged', 'value': 'b'},
     {'state': 'unchanged', 'value': 'c'},
     {'state': 'added', 'value': 'd'}]

    Parameters
    ----------
    before : iterable
        An iterable containing values to be used as the baseline version.
    after : iterable
        An iterable containing values to be compared against the baseline.
    check_modified : bool
        Whether or not to check for modifiedness.

    Returns
    -------
    diff_items : A list of dictionaries containing diff information.
    """

    # The grid will be empty if `before` or `after` are
    # empty; this will violate the assumptions made in the rest
    # of this function.
    # If this is the case, we know what the result of the diff is
    # anyways: the contents of the other, non-empty input.
    if len(before) == 0:
        return [
            {'state': 'added', 'value': v}
            for v in after
        ]
    elif len(after) == 0:
        return [
            {'state': 'deleted', 'value': v}
            for v in before
        ]

    grid = create_grid(before, after)

    nrows = len(grid[0])
    ncols = len(grid)
    dps = diff_points(grid)
    result = []
    for kind, col, row in dps:
        if kind == 'unchanged':
            value = before[col]
            result.append({
                'state': kind,
                'value': value,
            })
        elif kind == 'deleted':
            assert col < ncols
            value = before[col]
            result.append({
                'state': kind,
                'value': value,
            })
        elif kind == 'added':
            assert row < nrows
            value = after[row]
            result.append({
                'state': kind,
                'value': value,
            })
        elif check_modified and kind == 'modified':
            result.append({
                'state': kind,
                'originalvalue': before[col],
                'modifiedvalue': after[row],
            })
        elif (not check_modified) and kind == 'modified':
            result.append({
                'state': 'deleted',
                'value': before[col],
            })
            result.append({
                'state': 'added',
                'value': after[row],
            })
        else:
            raise Exception('We should not be here.')
    return result


def diff_points(grid):
    ncols = len(grid)
    nrows = len(grid[0])

    lcs_result = lcs(grid)
    matched_cols = [r[0] for r in lcs_result]
    matched_rows = [r[1] for r in lcs_result]

    cur_col = 0
    cur_row = 0

    result = []
    while cur_col < ncols or cur_row < nrows:
        passfirst = cur_col < ncols and cur_row < nrows
        goodrow = cur_row < nrows
        goodcol = cur_col < ncols
        if passfirst and lcs_result \
                and (cur_col, cur_row) == lcs_result[0]:
            lcs_result.pop(0)
            matched_cols.pop(0)
            matched_rows.pop(0)
            comparison = grid[cur_col][cur_row]
            if hasattr(comparison, 'is_modified') \
                    and comparison.is_modified():
                result.append(('modified', cur_col, cur_row))
            else:
                result.append(('unchanged', cur_col, cur_row))
            cur_col += 1
            cur_row += 1
        elif goodcol and \
                (not matched_cols or cur_col != matched_cols[0]):
            assert cur_col < ncols
            result.append(('deleted', cur_col, None))
            cur_col += 1
        elif goodrow and \
                (not matched_rows or cur_row != matched_rows[0]):
            assert cur_row < nrows
            result.append(('added', None, cur_row))
            cur_row += 1
#        print result

    return result


def create_grid(before, after):
    ncols = len(before)
    nrows = len(after)
    all_comps = [b == a for b, a in it.product(before, after)]
    return [
        all_comps[col*(nrows):col*(nrows)+nrows]
        for col in range(ncols)
    ]


def find_matches(col, colNum):
    result = []
    for j in range(len(col)):
        if col[j]:
            result.append((colNum, j))
    return result


def lcs(grid):
    kcs = find_candidates(grid)
    ks = kcs.keys()
    if len(ks) == 0:
        return []
    highest = max(kcs.keys())
    last_point = kcs[highest][-1]
    cur = highest - 1
    acc = [last_point]
    while cur > 0:
        comp = acc[-1]
        cx, cy = comp
        possibilities = [
            (x, y) for (x, y)
            in reversed(kcs[cur])
            if cx > x and cy > y
        ]
        if len(possibilities) > 0:
            acc.append(possibilities[-1])
        cur -= 1

    return list(reversed(acc))


def process_col(k, col, colNum):
    matches = find_matches(col, colNum)
    d = collections.defaultdict(lambda: [])
    x = 0
    for (i, j) in matches:
        oldx = x
        if not k and not d[1]:
            d[1].append((i, j))
        elif k:
            x = check_match((i, j), k)
            if x is None:
                continue
            x = x
            if x == oldx:
                continue
            d[x].append((i, j))
    return dict(d)


def check_match(point, k):
    result = []
    k_keys = k.keys()
    max_k = max(k_keys)
    new_max_k = max_k + 1
    k_range = k_keys + [new_max_k]
    for x in k_range:
        if x == 1:
            continue

        if point[1] < x-2:
            continue

        above_key = x - 1
        above_x = above_key == new_max_k and \
            10000 or max([l[0] for l in k[above_key]])
        above_y = above_key == new_max_k and \
            10000 or min([l[1] for l in k[above_key]])
        below_key = x - 2
        below_x = below_key < 1 and -1 or max([l[0] for l in k[below_key]])
        below_y = below_key < 1 and -1 or min([l[1] for l in k[below_key]])
        new_x, new_y = point
        if new_x > above_x and new_y < above_y and \
                new_x > below_x and new_y > below_y:
            result.append(x-1)

    below_key = new_max_k - 1
    below_x = below_key == 0 and -1 or max([l[0] for l in k[new_max_k-1]])
    below_y = below_key == 0 and -1 or min([l[1] for l in k[new_max_k-1]])
    new_x, new_y = point
    if new_x > below_x and new_y > below_y:
        result.append(new_max_k)
    if len(result) > 0:
        actual_result = result[0]
        # print result
        assert point[1] >= actual_result-1
        return (result)[0]
    else:
        return None


def add_results(k, result):
    finalResult = collections.defaultdict(lambda: [], k)
    for x in result.keys():
        finalResult[x] = finalResult[x] + result[x]
    return finalResult


def find_candidates(grid):
    k = collections.defaultdict(lambda: [])
    for colNum in range(len(grid)):
        k = add_results(k, process_col(k, grid[colNum], colNum))
    return dict(k)



from . import diff
from . import comparable
from .notebook_diff import (
    diff_result_to_cell,
)
import IPython.nbformat.current as nbformat
import itertools as it
import copy


def merge(local, base, remote, check_modified=False):
    """Generate unmerged series of changes (including conflicts).

    By diffing the two diffs, we find *changes* that are
    on the local branch, the remote branch, or both.
    We arbitrarily choose the "local" branch to be the "before"
    and the "remote" branch to be the "after" in the diff algorithm.

    Therefore:
    If a change is "deleted", that means that it occurs only on
    the local branch. If a change is "added" that means it occurs only on
    the remote branch. If a change is "unchanged", that means it occurs
    in both branches. Either the same addition or same deletion occurred in
    both branches, or the cell was not changed in either branch.

    Parameters
    ----------
    local : list
        A sequence representing the items on the local branch.
    base : dict
        A sequence representing the items on the base branch
    remote : dict
        A sequence representing the items on the remote branch.

    Returns
    -------
    result : A diff result comparing the changes on the local and remote
             branches.
    """
    base_local = diff.diff(base, local, check_modified=check_modified)
    base_remote = diff.diff(base, remote, check_modified=check_modified)
    merge = diff.diff(base_local, base_remote)
    return merge


def notebook_merge(local, base, remote, check_modified=False):
    """Unify three notebooks into a single notebook with merge metadata.

    The result of this function is a valid notebook that can be loaded
    by the IPython Notebook front-end. This function adds additional
    cell metadata that the front-end Javascript uses to render the merge.

    Parameters
    ----------
    local : dict
        The local branch's version of the notebook.
    base : dict
        The last common ancestor of local and remote.
    remote : dict
        The remote branch's version of the notebook.

    Returns
    -------
    nb : A valid notebook containing merge metadata.
    """

    local_cells = get_cells(local)
    base_cells = get_cells(base)
    remote_cells = get_cells(remote)

    rows = []
    current_row = []
    empty_cell = lambda: {
        'cell_type': 'code',
        'language': 'python',
        'outputs': [],
        'prompt_number': 1,
        'text': ['Placeholder'],
        'metadata': {'state': 'empty'}
    }

    diff_of_diffs = merge(local_cells, base_cells, remote_cells)

    # For each item in the higher-order diff, create a "row" that
    # corresponds to a row in the NBDiff interface. A row contains:
    # | LOCAL | BASE | REMOTE |

    for item in diff_of_diffs:
        state = item['state']
        cell = copy.deepcopy(diff_result_to_cell(item['value']))
        if state == 'deleted':
            # This change is between base and local branches.
            # It can be an addition or a deletion.
            if cell['metadata']['state'] == 'unchanged':
                # This side doesn't have the change; wait
                # until we encounter the change to create the row.
                continue
            cell['metadata']['side'] = 'local'
            remote_cell = empty_cell()
            remote_cell['metadata']['side'] = 'remote'
            if cell['metadata']['state'] == 'deleted' \
                    or cell['metadata']['state'] == 'unchanged':
                base_cell = copy.deepcopy(cell)
            else:
                base_cell = empty_cell()
            base_cell['metadata']['side'] = 'base'
            # This change is on the right.
            current_row = [
                cell,
                base_cell,
                remote_cell,
            ]
        elif state == 'added':
            # This change is between base and remote branches.
            # It can be an addition or a deletion.
            cell['metadata']['side'] = 'remote'
            if cell['metadata']['state'] == 'unchanged':
                # This side doesn't have the change; wait
                # until we encounter the change to create the row.
                continue
            if cell['metadata']['state'] == 'deleted':
                base_cell = copy.deepcopy(cell)
                base_cell['metadata']['state'] = 'unchanged'
                local_cell = copy.deepcopy(cell)
                local_cell['metadata']['state'] = 'unchanged'
            else:
                base_cell = empty_cell()
                local_cell = empty_cell()
            base_cell['metadata']['side'] = 'base'
            local_cell['metadata']['side'] = 'local'
            current_row = [
                local_cell,
                base_cell,
                cell,
            ]
        elif state == 'unchanged':
            # The same item occurs between base-local and base-remote.
            # This happens if both branches made the same change, whether
            # that is an addition or deletion. If neither branches
            # changed a given cell, that cell shows up here too.
            cell1 = copy.deepcopy(cell)
            cell3 = copy.deepcopy(cell)
            if cell['metadata']['state'] == 'deleted' \
                    or cell['metadata']['state'] == 'unchanged':
                # If the change is a deletion, the cell-to-be-deleted
                # should in the base as 'unchanged'. The user will
                # choose to make it deleted.
                cell2 = copy.deepcopy(cell)
                cell2['metadata']['state'] = 'unchanged'
            else:
                # If the change is an addition, it should not
                # show in the base; the user must add it to the merged version.
                cell2 = empty_cell()
            cell1['metadata']['side'] = 'local'
            cell2['metadata']['side'] = 'base'
            cell3['metadata']['side'] = 'remote'
            current_row = [
                cell1,
                cell2,
                cell3,
            ]

        rows.append(current_row)

    # Chain all rows together; create a flat array from the nested array.
    # Use the base notebook's notebook-level metadata (title, version, etc.)

    result_notebook = local
    if len(result_notebook['worksheets']) == 0:
        result_notebook['worksheets'] = [nbformat.new_worksheet()]

    new_cell_array = list(it.chain.from_iterable(rows))
    result_notebook['worksheets'][0]['cells'] = new_cell_array

    result_notebook['metadata']['nbdiff-type'] = 'merge'

    return result_notebook


def get_cells(notebook, check_modified=False):
    try:
        cells = [
            comparable.CellComparator(cell, check_modified=check_modified)
            for cell in
            notebook['worksheets'][0]['cells']
        ]
    except IndexError:
        cells = []
    except KeyError:
        cells = []
    return cells



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import os
import sys


try:
    from setuptools import setup
except ImportError:
    from distutils.core import setup

if sys.argv[-1] == 'publish':
    os.system('python setup.py sdist upload')
    sys.exit()

readme = open('README.rst').read()
history = open('HISTORY.rst').read().replace('.. :changelog:', '')

setup(
    name='nbdiff',
    version='1.0.0',
    description='A tool for diffing and merging IPython Notebook files',
    long_description=readme + '\n\n' + history,
    author='Tavish Armstrong',
    author_email='tavisharmstrong@gmail.com',
    url='https://github.com/tarmstrong/nbdiff',
    packages=[
        'nbdiff',
        'nbdiff.server',
        'nbdiff.adapter',
    ],
    package_data={
        'nbdiff.server': [
            'templates/*',
            'static/libraries/*',
            'static/image/*',
            'static/js/*',
            'static/css/*',
        ],
    },
    package_dir={'nbdiff': 'nbdiff'},
    entry_points={'console_scripts': [
        'nbdiff = nbdiff.commands:diff',
        'nbmerge = nbdiff.commands:merge',
    ]},
    include_package_data=True,
    install_requires=[
        'flask',
        'jinja2',
        'ipython',
        'python-Levenshtein',
    ],
    license="MIT",
    zip_safe=False,
    keywords='nbdiff',
    classifiers=[
        'Development Status :: 2 - Pre-Alpha',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: BSD License',
        'Natural Language :: English',
        "Programming Language :: Python :: 2",
        'Programming Language :: Python :: 2.6',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.3',
    ],
    test_suite='tests',
)



from flask import Flask, request, render_template
from nbdiff.server.database import db_session
from nbdiff.server.database import init_db
import sys
import jinja2
import IPython.html
import os


# initialize database if argument 'True' is passed
if len(sys.argv) > 1:
    if(sys.argv[1].lower() == "init_db"):
        init_db()


class RemoteNbFlask(Flask):
    jinja_loader = jinja2.FileSystemLoader([
        IPython.html.__path__[0] + '/templates',
        os.path.dirname(os.path.realpath(__file__)) + '/templates'
    ])

    def shutdown_callback(self, callback):
        self.shutdown = callback
        db_session.remove()

app = RemoteNbFlask(
    __name__,
    static_folder=IPython.html.__path__[0] + '/static'
)


def get_class(classname):
    components = classname.split('.')
    try:
        obj = __import__(classname)
    except ImportError:
        raise ImportError

    for comp in components[1:]:
        obj = getattr(obj, comp)
    return obj


def run_command(cmdName, request, filename=None):
    cmd = "nbdiff.server.command."+cmdName+"Command"
    try:
        command = get_class(cmd).newInstance()
    except ImportError:
        errMsg = "404: The page requested does not Exist!"
        return render_template('Error.html', err=errMsg)
    return command.process(request, filename, db_session())


# index
@app.route("/")
def upload():
    return run_command("Upload", request)


# runs depending on different command URL
@app.route("/<path:command>", methods=['GET', 'POST'])
def redirectCommand(command):
    # favicon.ico is a resource requested by Ipython notebook.
    # Since it does not follow command pattern this will
    # redirect to proper command
    if command == "favicon.ico":
        url = "image/favicon.ico"
        return run_command("ResourceRequest", request, url)
    else:
        return run_command(command, request)


# notebook request handler
@app.route('/notebooks/<path:filename>', methods=['GET', 'PUT'])
def notebookRequest(filename):
    return run_command("NotebookRequest", request, filename)


# used to get specific resources in the html pages.
@app.route('/nbdiff/<path:filename>')
def nbdiff_static(filename):
    return run_command("ResourceRequest", request, filename)


# Redirect from Merge, MergeURL, Diff, DiffURL commands.
# Used to simplify URL for users who wish to go back to their comparison.
@app.route('/Comparison/<path:filename>')
def comparisonURL(filename):
    return run_command("Comparison", request, filename)

if __name__ == "__main__":
    app.debug = False
    app.run()



from flask import Flask, render_template, send_from_directory, request
import jinja2
import json
import IPython.html
import os


class NbFlask(Flask):
    jinja_loader = jinja2.FileSystemLoader([
        IPython.html.__path__[0] + '/templates',
        os.path.dirname(os.path.realpath(__file__)) + '/templates'
    ])

    notebooks = []

    def shutdown_callback(self, callback):
        self.shutdown = callback

    def add_notebook(self, nb, fname):
        self.notebooks.append((nb, fname))

app = NbFlask(__name__, static_folder=IPython.html.__path__[0] + '/static')


@app.route('/nbdiff/<path:filename>')
def nbdiff_static(filename):
    return send_from_directory(os.path.dirname(os.path.realpath(__file__))
                               + '/static', filename)


@app.route('/<int:notebookid>')
def home(notebookid):
    return render_template(
        'nbdiff.html',
        project='/',
        base_project_url='/',
        base_kernel_url='/',
        notebook_id='test_notebook' + str(notebookid),
        num_nbks=str(len(app.notebooks)),
        cur_nbk=str(notebookid),
        local=True,
    )


@app.route('/notebooks/test_notebook<int:notebookid>', methods=['GET', 'PUT'])
def notebookjson(notebookid):
    if request.method == 'PUT':
        app.shutdown(request.data, app.notebooks[notebookid][1])
        return ""
    else:
        parsed, filename = app.notebooks[notebookid]
        parsed['metadata']['filename'] = filename
        return json.dumps(parsed)


@app.route('/shutdown')
def shutdown():
    request.environ.get('werkzeug.server.shutdown')()
    return "The server was shutdown."

if __name__ == '__main__':
    app.run(debug=True)



from .diff import diff
from .comparable import CellComparator, LineComparator


def notebook_diff(nb1, nb2, check_modified=True):
    """Unify two notebooks into a single notebook with diff metadata.

    The result of this function is a valid notebook that can be loaded
    by the IPython Notebook front-end. This function adds additional
    cell metadata that the front-end Javascript uses to render the diffs.

    Parameters
    ----------
    nb1 : dict
        An IPython Notebook to use as the baseline version.
    nb2 : dict
        An IPython Notebook to compare against the baseline.
    check_modified : bool
        Whether or not to detect cell modification.

    Returns
    -------
    nb : A valid notebook containing diff metadata.
    """
    nb1_cells = nb1['worksheets'][0]['cells']
    nb2_cells = nb2['worksheets'][0]['cells']

    diffed_nb = cells_diff(nb1_cells, nb2_cells, check_modified=check_modified)
    line_diffs = diff_modified_items(diffed_nb)

    cell_list = list()
    for i, item in enumerate(diffed_nb):
        cell = diff_result_to_cell(item)
        if i in line_diffs:
            cell['metadata']['extra-diff-data'] = line_diffs[i]
        cell_list.append(cell)

    nb1['worksheets'][0]['cells'] = cell_list
    nb1['metadata']['nbdiff-type'] = 'diff'

    return nb1


def diff_modified_items(cellslist):
    result = {}
    for i in range(len(cellslist)):
        if cellslist[i]['state'] == 'modified':
            if cellslist[i]['originalvalue'].data['cell_type'] == 'heading':
                result[i] = diff(
                    cellslist[i]['originalvalue'].data["source"].split(),
                    cellslist[i]['modifiedvalue'].data["source"].split(),
                )
            else:
                result[i] = diff(
                    cellslist[i]['originalvalue'].data["input"].splitlines(),
                    cellslist[i]['modifiedvalue'].data["input"].splitlines(),
                )
    return result


def diff_result_to_cell(item):
    '''diff.diff returns a dictionary with all the information we need,
    but we want to extract the cell and change its metadata.'''
    state = item['state']
    if state == 'modified':
        new_cell = item['modifiedvalue'].data
        old_cell = item['originalvalue'].data
        new_cell['metadata']['state'] = state
        new_cell['metadata']['original'] = old_cell
        cell = new_cell
    else:
        cell = item['value'].data
        cell['metadata']['state'] = state
    return cell


def cells_diff(before_cells, after_cells, check_modified=False):
    '''Diff two arrays of cells.'''
    before_comps = [
        CellComparator(cell, check_modified=check_modified)
        for cell in before_cells
    ]
    after_comps = [
        CellComparator(cell, check_modified=check_modified)
        for cell in after_cells
    ]
    diff_result = diff(
        before_comps,
        after_comps,
        check_modified=check_modified
    )
    return diff_result


def words_diff(before_words, after_words):
    '''Diff the words in two strings.

    This is intended for use in diffing prose and other forms of text
    where line breaks have little semantic value.

    Parameters
    ----------
    before_words : str
        A string to be used as the baseline version.
    after_words : str
        A string to be compared against the baseline.

    Returns
    -------
    diff_result : A list of dictionaries containing diff information.
    '''
    before_comps = before_words.split()
    after_comps = after_words.split()

    diff_result = diff(
        before_comps,
        after_comps
    )
    return diff_result


def lines_diff(before_lines, after_lines, check_modified=False):
    '''Diff the lines in two strings.

    Parameters
    ----------
    before_lines : iterable
        Iterable containing lines used as the baseline version.
    after_lines : iterable
        Iterable containing lines to be compared against the baseline.

    Returns
    -------
    diff_result : A list of dictionaries containing diff information.
    '''
    before_comps = [
        LineComparator(line, check_modified=check_modified)
        for line in before_lines
    ]
    after_comps = [
        LineComparator(line, check_modified=check_modified)
        for line in after_lines
    ]
    diff_result = diff(
        before_comps,
        after_comps,
        check_modified=check_modified
    )
    return diff_result
